/*-------------------------------------------------------------------
  All code is provided for EDUCATIONAL purposes ONLY. There is no
  warranty and no support.

  ds1775.c - DS1775R I2C sensor support functions

  Processor: ATtiny854A
      Clock: 8 mHz internal

  2019-02-03 written
-------------------------------------------------------------------*/
#include <avr/io.h>
#include <util/delay_basic.h>
#include <stdlib.h>
#include <string.h>
#include "main.h"
#include "ds1775.h"
#include "i2cmaster.h"

/*-----------------------------------------------------------------
  ds1775_write_control_register(...)

  Input: control register

  returns: 0 = OK
           1 = START failed to access device
           2 = write pointer byte fail
           3 = write control byte fail
----------------------------------------------------------------*/
unsigned char ds1775_write_config_register( unsigned char ucConfigReg )
{
  unsigned char ucRc;

  // Access the bus and device, output the address
  ucRc = i2c_start( DS1775_ADDRESS_W );
  if ( ucRc == 1 )             // fail to access device
  {
    return 1;
  }

  // Output the register pointer (config byte register)
  ucRc = i2c_write( DS1775_CONFIG_REG );
  if ( ucRc == 1 )             // fail
  {
    i2c_stop();
    return 2;
  }

  // write the configuration register
  ucRc = i2c_write( ucConfigReg );
  i2c_stop();
  if ( ucRc == 1 )             // fail
  {
    return 3;
  }

  // Select the temperature register for future access

  ucRc = i2c_start( DS1775_ADDRESS_W );
  if ( ucRc == 1 )             // fail to access device
  {
    return 1;
  }

  // Output the register pointer (temperature register)
  ucRc = i2c_write( DS1775_TEMP_REG );
  i2c_stop();
  if ( ucRc == 1 )             // fail
  {
    return 2;
  }

  return 0;
}

/*-----------------------------------------------------------------
  ds1775_read_temperature(...)

  In/out: pointer to signed integer to receive temperature value...
          This is the raw 2-byte signed integer value returned 
          by the device.

  returns: 0 = OK
           1 = START failed to access device
----------------------------------------------------------------*/
unsigned char ds1775_read_temperature( volatile int *piTemperature )
{
  unsigned char ucRc;
  unsigned int  uiTemp;

  // Access the bus and device, output the address
  ucRc = i2c_start( DS1775_ADDRESS_R );
  if ( ucRc == 1 )             // fail to access device
  {
    return 1;
  }

  // Read temperature MSB
  ucRc = i2c_readAck();
  uiTemp = (unsigned int) ucRc << 8;

  // Read temperature LSB
  ucRc = i2c_readNak();
  i2c_stop();

  uiTemp = uiTemp | (unsigned int) ucRc;
  *piTemperature = (int) uiTemp;

  return 0;
}

/*-----------------------------------------------------------------
  ds1775_write_temperature(...)

  Input: iTos - raw temperature to put in Tos register
         iThyst - raw temperature to put in Thyst register.

  returns: 0 = OK
           1 = START failed to access device
           2 = write pointer byte fail
           3 = write control byte fail
----------------------------------------------------------------*/
unsigned char ds1775_write_temperature( int iTos, int iThyst )
{
  unsigned char ucRc, ucByte;

  // Access the bus and device, output the address
  ucRc = i2c_start( DS1775_ADDRESS_W );
  if ( ucRc == 1 )             // fail to access device
  {
    return 1;
  }

  // Output the register pointer (config byte register)
  ucRc = i2c_write( DS1775_TOS_REG );
  if ( ucRc == 1 )             // fail
  {
    i2c_stop();
    return 2;
  }

  // write the MSB
  ucByte = (unsigned char) ( (unsigned int) iTos >> 8 );
  ucRc = i2c_write( ucByte );
  if ( ucRc == 1 )             // fail
  {
    i2c_stop();
    return 3;
  }

  // write the LSB
  ucByte = (unsigned char) ( (unsigned int) iTos & 0x00ff );
  ucRc = i2c_write( ucByte );
  i2c_stop();
  if ( ucRc == 1 )             // fail
  {
    return 3;
  }

  // Access the bus and device, output the address
  ucRc = i2c_start( DS1775_ADDRESS_W );
  if ( ucRc == 1 )             // fail to access device
  {
    return 1;
  }

  // Output the register pointer (config byte register)
  ucRc = i2c_write( DS1775_THYST_REG );
  if ( ucRc == 1 )             // fail
  {
    i2c_stop();
    return 2;
  }

  // write the MSB
  ucByte = (unsigned char) ( (unsigned int) iThyst >> 8 );
  ucRc = i2c_write( ucByte );
  if ( ucRc == 1 )             // fail
  {
    i2c_stop();
    return 3;
  }

  // write the LSB
  ucByte = (unsigned char) ( (unsigned int) iThyst & 0x00ff );
  ucRc = i2c_write( ucByte );
  i2c_stop();
  if ( ucRc == 1 )             // fail
  {
    return 3;
  }

  // Select the temperature register for future access

  ucRc = i2c_start( DS1775_ADDRESS_W );
  if ( ucRc == 1 )             // fail to access device
  {
    return 1;
  }

  // Output the register pointer (temperature register)
  ucRc = i2c_write( DS1775_TEMP_REG );
  i2c_stop();
  if ( ucRc == 1 )             // fail
  {
    return 2;
  }

  return 0;
}

/*-----------------------------------------------------------------
  ds1775_raw_to_tenths(...)

  Converts the raw temperature signed integer from the device to 
  tenths degrees C.

  Input:  temperature value - this is the raw 2-byte signed 
                              integer value from the device.

  returns: temperature in 10the degrees C
----------------------------------------------------------------*/
int ds1775_raw_to_tenths( int iRawTemp )
{
  int iTemp10thsC;

  iTemp10thsC = (int) ( ( (long) iRawTemp * 10L ) / 256L );

  return iTemp10thsC;
}

/*-----------------------------------------------------------------
  ds1775_tenths_to_raw(...)

  Converts tenths degrees C to the raw temperature signed integer
  for the device.

  Input:  temperature value - temperature in 10ths degrees C

  returns: raw 2-byte signed integer value for the device.
----------------------------------------------------------------*/
int ds1775_tenths_to_raw( int iTemp10thsC )
{
  int iTempRaw;

  iTempRaw = (int) ( ( (long) iTemp10thsC * 256L ) / 10L );

  return iTempRaw;
}
