/*-----------------------------------------------------------------
  All code is provided for EDUCATIONAL purposes ONLY. There is no
  warranty and no support.

  Copyright (c) Jack Botner and licensed under the MIT license; see
  http://opensource.org/licenses/mit-license.php

  space heater controller by Jack Botner

  space heater controller(2) - main.c

  This program implements a thermostat for controlling a space heater.
  The LED display shows both room temperature and set temperature.
  - two 2-digit LED displays LTD-4608JF
  - SAA1064 I2C LED controller
  - a CTS rotary encoder with pushbutton
  - a relay that controls the heater
  - DS1775 I2C temperature sensor

  Processor: ATtiny84A
  Clock: Internal, 8.0 MHz

  20190210 program written
-----------------------------------------------------------------*/
#define F_CPU 8000000UL			// 8 MHz
#include <stdint.h>
#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <util/delay_basic.h>
#include <stdlib.h>
#include <string.h>
#include "main.h"
#include "saa1064.h"
#include "ds1775.h"
#include "i2cmaster.h"

#undef WANT_WDT

// Global Variables

volatile uint8_t ucEventFlags=0;		// program event flags

static volatile int iMeasTempC=210;	    // measured temperature (10ths degrees C)

static MY_EEPROM_DATA stEepromVars;	    // variables kept in EEPROM

// Local functions

static void blink_diagnostic_led( unsigned char ucNumber );
static void display_temperature( unsigned char ucWhichDisplay, int iTemperature );
static void project_initialization(void);
static void system_init(void);

/*-----------------------------------------------------------------
  main()
-----------------------------------------------------------------*/
int main(void)
{
  unsigned char ucRc, ucCtr=5; // , ucOn=0;
  int           iTempRaw;
  int           iTos, iThyst, iTemp;

  // WDT: There is an issue here because of the use of the I2C devices. As long as
  // everything is working, then no problem. However if an error occurs, there is
  // little error handling or correction. Basically the program hangs in
  // blink_diagnostic_led(). If the WDT is running, this results in a reset and
  // the program restarts. If the error happens to be transient, great. However
  // it is very likely that whatever caused the error will happen again. For example
  // someone unplugs the temperature sensor, or a wire breaks, etc. This puts the
  // project in a restart loop with no possibility of resolution. We need to simply
  // shut off the load and call it a day. So it may be best not to use the WDT at all.
#ifdef WANT_WDT
  int           ii;
  uint8_t       mcusr_mirror;

  // Following code is supposed to detect if a watchdog reset occurred followed
  // by a restart. Since the load is likely a space heater, we want to ensure
  // it is turned off and stays off for awhile in case a repeating condition occurs.
  mcusr_mirror = MCUSR;
  MCUSR = 0;
  wdt_disable();   // libc says to do this now
#endif

  project_initialization();

#ifdef WANT_WDT
  if ( test_bit( mcusr_mirror, WDRF ) )
  {
    clear_bit( RELAY_PORT_OUT, RELAY_PIN );    // make sure load is off
	set_bit( DIAG_LED_PORT, DIAG_LED_PIN );
	for ( ii=0; ii<18321; ++ii )               // 10 minutes
	  _delay_loop_2( 0 ); 
	clear_bit( DIAG_LED_PORT, DIAG_LED_PIN );
  }
#endif
  
  if ( stEepromVars.ucSystemOn )   // show set temperature
    display_temperature( DISPLAY_1, stEepromVars.iSetTempC );

  while (1)
  {
    if ( test_bit( ucEventFlags, EVENT_ANALYZE_STATE ) )
    {
      // Do this every 6 seconds
      ++ucCtr;
      if ( ucCtr > 6 )
	  // Note there is an oddity here ... when powered on, for a few seconds
	  // the OS pin is asserted even if it should be in the off state. Eventually
	  // it goes off. When I was testing for 5 seconds the load would come on
	  // (1st pass through here), then 5 seconds later turn off. Does not happen 
	  // when delay is 6 seconds.
      {
        // Measure and display the temperature
        ucRc = ds1775_read_temperature( &iTempRaw );
        if ( ucRc )
          blink_diagnostic_led( ucRc );

        iMeasTempC = ds1775_raw_to_tenths( iTempRaw );
        display_temperature( DISPLAY_2, iMeasTempC );

        // Test the OS pin
        if ( stEepromVars.ucSystemOn )
        {
          if ( test_bit( OS_PORT_IN, OS_PIN ) )
          {
            set_bit( RELAY_PORT_OUT, RELAY_PIN );
            //set_bit( DIAG_LED_PORT, DIAG_LED_PIN );
          }
          else
          {
            clear_bit( RELAY_PORT_OUT, RELAY_PIN );
            //clear_bit( DIAG_LED_PORT, DIAG_LED_PIN );
          }
        }
        else    // power is off
        {
          clear_bit( RELAY_PORT_OUT, RELAY_PIN );
          //clear_bit( DIAG_LED_PORT, DIAG_LED_PIN );
        }

        ucCtr = 0;
      }

      clear_bit( ucEventFlags, EVENT_ANALYZE_STATE );
    }

    if ( test_bit( ucEventFlags, EVENT_ENCODER_UP ) )
    {
      if ( stEepromVars.ucSystemOn )
      {
        // Increment the set temperature, limited to 40C
        if ( stEepromVars.iSetTempC < 390 )
          stEepromVars.iSetTempC += 10;

        display_temperature( DISPLAY_1, stEepromVars.iSetTempC );
        set_bit( ucEventFlags, EVENT_UPDATE_TOS_THYST );
        set_bit( ucEventFlags, EVENT_UPDATE_EEPROM );
      }

      clear_bit( ucEventFlags, EVENT_ENCODER_UP );
    }

    if ( test_bit( ucEventFlags, EVENT_ENCODER_DOWN ) )
    {
      if ( stEepromVars.ucSystemOn )
      {
        // Decrement the set temperature, limited to 1C
        if ( stEepromVars.iSetTempC > 10 )
          stEepromVars.iSetTempC -= 10;

        display_temperature( DISPLAY_1, stEepromVars.iSetTempC );
        set_bit( ucEventFlags, EVENT_UPDATE_TOS_THYST );
        set_bit( ucEventFlags, EVENT_UPDATE_EEPROM );
      }

      clear_bit( ucEventFlags, EVENT_ENCODER_DOWN );
    }

    if ( test_bit( ucEventFlags, EVENT_BUTTON_PRESSED ) )
    {
      // Turn the unit on or off
      if ( stEepromVars.ucSystemOn )
      {
        // Turn off the set temperature display to show unit is off
        ucRc = saa1064_write_digit( SAA1064_DIGIT_1_PTR, UNTRANSLATED_MODE, 0 );
        ucRc = saa1064_write_digit( SAA1064_DIGIT_2_PTR, UNTRANSLATED_MODE, 0 );
        stEepromVars.ucSystemOn = 0;
      }
      else
      {
        display_temperature( DISPLAY_1, stEepromVars.iSetTempC );
        stEepromVars.ucSystemOn = 1;
      }

      set_bit( ucEventFlags, EVENT_UPDATE_EEPROM );
      clear_bit( ucEventFlags, EVENT_BUTTON_PRESSED );
    }

    if ( test_bit( ucEventFlags, EVENT_UPDATE_EEPROM ) )
    {
      eeprom_write_block( &stEepromVars, 0, sizeof( MY_EEPROM_DATA ) );
      clear_bit( ucEventFlags, EVENT_UPDATE_EEPROM );
    }

    if ( test_bit( ucEventFlags, EVENT_UPDATE_TOS_THYST ) )
    {
      // Write Tos and Thyst
      iTemp = stEepromVars.iSetTempC + ( TEMP_HYSTERESIS_TENTHS / 2 );
      iTos = ds1775_tenths_to_raw( iTemp );
      iTemp = stEepromVars.iSetTempC - ( TEMP_HYSTERESIS_TENTHS / 2 );
      iThyst = ds1775_tenths_to_raw( iTemp );
      ucRc = ds1775_write_temperature( iTos, iThyst );
      if ( ucRc )
        blink_diagnostic_led( ucRc );

      clear_bit( ucEventFlags, EVENT_UPDATE_TOS_THYST );
    }

#ifdef WANT_WDT
    wdt_reset();
#endif	
  }
}

/*-----------------------------------------------------------------
  project_initialization()

  Initialize the microcontroller
  Initialize the display driver
  Show initial display
  Initialize the temperature sensor
----------------------------------------------------------------*/
static void project_initialization()
{
  unsigned char ucRc;
  char ii;

  system_init();                	// initialize system

  // Turn on all segments for a second or two
  ucRc = saa1064_write_control_register( 0b00011111 );
  if ( ucRc )
    blink_diagnostic_led( ucRc );
  
  // Get the saved state variables from eeprom
  eeprom_read_block( &stEepromVars, 0, sizeof( MY_EEPROM_DATA ) );
  if ( stEepromVars.iSetTempC > 400 || stEepromVars.iSetTempC < 10 ||
       stEepromVars.ucSystemOn > 1 )		   // probably EEPROM not initialized
  {
    stEepromVars.iSetTempC = 210;
    stEepromVars.ucSystemOn = 0;
    set_bit( ucEventFlags, EVENT_UPDATE_EEPROM );
  }

  for ( ii=0; ii<60; ++ii )			// 2 sec delay
    _delay_loop_2( 65000 );			// 32 ms delay

  // Turn on all segments off, ready for data
  ucRc = saa1064_write_control_register( 0b00010111 );
  if ( ucRc )
    blink_diagnostic_led( ucRc );

  for ( ii=0; ii<4; ++ii )
  {
    ucRc = saa1064_write_digit( ( ii + SAA1064_DIGIT_1_PTR ), UNTRANSLATED_MODE, 0 );
    if ( ucRc )
      blink_diagnostic_led( ucRc );
  }

  // Initialize the DS1775R temperature sensor
  // Note: I have set the POLarity bit contrary to the documentation;
  // either the documentation is wrong or I have a logic bug I cannot find.
  ucRc = ds1775_write_config_register( 0b00001000 );
  if ( ucRc )
    blink_diagnostic_led( ucRc );

  set_bit( ucEventFlags, EVENT_UPDATE_TOS_THYST );
#ifdef WANT_WDT
  wdt_enable( WDTO_4S );    // use the WDT to prevent lockup with load on
#endif  
}

/*-----------------------------------------------------------------
  system_init()

  PORTA: PA0 (input)  Encoder A
		     PA1 (input)  Encoder B
		     PA2 (input)  Encoder SW
		     PA3 (output) Relay
		     PA4 (input)  programming
		     PA5 (input)  programming
		     PA6 (input)  programming
		     PA7 (input)  DS1775 Thermostat Output

  PORTB: PB0 (inout)  SDA
         PB1 (output) SCL
         PB2 (input)  not used
         PB3 (input)  programming
----------------------------------------------------------------*/
static void system_init()
{
  // The fuses are set to provide an internal clock, running at
  // 8.0 MHz with a prescaler of 8. Change the prescaler from
  // 8 to 1. The procedure requires two steps as per the device
  // specifications.
  CLKPR = 0b10000000;       // set clock prescaler change enable
  CLKPR = 0b00000000;       // set division factor to 1

  // Set up the I/O ports directions, pull-ups
  DDRA  = 0b00001000;
  PORTA = 0b11110111;       // initialize PortA

  DDRB  = 0b00000111;
  PORTB = 0b00001000;       // initialize PortB

  // Initialize I2C
  i2c_init();

  // Initialize Timer0. Timer0 is used to generate interrupts
  // every 1 millisecond for switch debouncing.
  TCCR0A = 0b00000000;		// normal port operation
  TCCR0B = 0b00000011;		// normal operation, prescaler=64
  TCNT0 = INIT_TIMER0;
  set_bit( TIMSK0, TOIE0 );	// enable timer/counter2 overflow interrupt

  // Initialize Timer1.
  // Timer1 is used to create 1 s intervals for scheduling events
  TCCR1A = 0b00000000;		  // normal operation, no OC1A, no OC1B
  TCCR1B = 0b00000100;      // internal clock, prescaler = 256
  TCNT1 = INIT_TIMER1;
  set_bit( TIMSK1, TOIE1 );	// enable timer/counter1 overflow interrupt

  PRR = 0b00000011;			  // power save: no USI, no ADC
  sei();                  // enable global interrupts
}

/*-----------------------------------------------------------------
  display_temperature(...)

  ucWhichDisplay = 0 (left display) or 1 (right display)
  iTemperature = temperature in 10ths degrees C
----------------------------------------------------------------*/
static void display_temperature( unsigned char ucWhichDisplay, int iTemperature )
{
  unsigned char ucRc;
  int           iDividend, iRemainder, iShow;

  iRemainder = iTemperature % 10;     // round the value
  iDividend = iTemperature / 10;
  iShow = ( iRemainder > 4 ) ? ( iDividend + 1 ) : iDividend;
  ucRc = saa1064_write_number( ucWhichDisplay, iShow );
  if ( ucRc )   // error
    blink_diagnostic_led( ucRc );
}

/*-----------------------------------------------------------------
  blink_diagnostic_led(...)

  Display a diagnostic number by blinking the diagnostic LED
  forever.

  Exception: number = 0, blink LED once and return.
----------------------------------------------------------------*/
static void blink_diagnostic_led( unsigned char ucNumber )
{
  unsigned char ii, jj;

  if ( !ucNumber )
  {
    set_bit( PORTB, PB2 );        // just turn LED on
    for ( ii=0; ii<15; ++ii )     // 1/2 sec delay
    _delay_loop_2( 65000 );		  // 32 ms delay
    clear_bit( PORTB, PB2 );      // turn LED off
    return;
  }

  // make sure load is off now
  clear_bit( RELAY_PORT_OUT, RELAY_PIN );

  while (1)
  {
    for ( jj=0; jj<ucNumber; ++jj )
    {
      set_bit( PORTB, PB2 );        // turn LED on
      for ( ii=0; ii<15; ++ii )			// 1/2 sec delay
        _delay_loop_2( 65000 );			// 32 ms delay
      clear_bit( PORTB, PB2 );      // turn LED off
      for ( ii=0; ii<15; ++ii )			// 1/2 sec delay
      _delay_loop_2( 65000 );			  // 32 ms delay
    }

    for ( ii=0; ii<45; ++ii )			  // 1.5 sec delay
      _delay_loop_2( 65000 );			  // 32 ms delay
  }
}