/*-------------------------------------------------------------------
  All code is provided for EDUCATIONAL purposes ONLY. There is no
  warranty and no support.

  saa1064.c - SAA1064 I2C sensor support functions

  Processor: ATtiny854A
      Clock: 8 mHz internal

  2019-02-03 written
-------------------------------------------------------------------*/
#include <avr/io.h>
#include <util/delay_basic.h>
#include <stdlib.h>
#include <string.h>
#include "main.h"
#include "saa1064.h"
#include "i2cmaster.h"

static unsigned char translate( unsigned char ucValue );

/*-----------------------------------------------------------------
  saa1064_write_control_register(...)

  Input: control register

  returns: 0 = OK
           1 = START failed to access device
           2 = write pointer byte fail
           3 = write control byte fail
----------------------------------------------------------------*/
unsigned char saa1064_write_control_register( unsigned char ucControlReg )
{
  unsigned char ucRc;

  // Access the bus and device, output the address
  ucRc = i2c_start( SAA1064_ADDRESS_W );
  if ( ucRc == 1 )             // fail to access device
  {
    return 1;
  }

  // Output the register pointer (control byte register)
  ucRc = i2c_write( SAA1064_CONTROL_PTR );
  if ( ucRc == 1 )             // fail
  {
    i2c_stop();
    return 2;
  }

  // write the configuration register
  ucRc = i2c_write( ucControlReg );
  i2c_stop();
  if ( ucRc == 1 )             // fail
  {
    return 3;
  }

  return 0;
}

/*-----------------------------------------------------------------
  saa1064_write_number(...)

  Input: which display
           0 = left side display
           1 = right side display
         number to display
           signed integer with the values -9 to 99.

  returns: 0 = OK
           1 = START failed to access device
           2 = write pointer byte fail
           3 = write digit fail
           4 = bad parameter
----------------------------------------------------------------*/
unsigned char saa1064_write_number( unsigned char ucWhichDisplay, int iNumber )
{
  unsigned char ucRc;
  unsigned char ucLow, ucHigh;
  unsigned char ucDigit1, ucDigit2;
  
  if ( iNumber < -9 )
    return 4;
  
  if ( iNumber > 99 )
    return 4;

  if ( !ucWhichDisplay )
  {
    ucDigit1 = SAA1064_DIGIT_1_PTR;
    ucDigit2 = SAA1064_DIGIT_2_PTR;
  }
  else
    if ( ucWhichDisplay == 1 )
    {
      ucDigit1 = SAA1064_DIGIT_3_PTR;
      ucDigit2 = SAA1064_DIGIT_4_PTR;
    }
    else
      return 4;

  // We have a number between -1 and 99. We need two digits, a high digit
  // or minus sign, and a low digit
 
  if ( iNumber < 0 )
  {
    // Display a minus sign on digit 1
    ucRc = saa1064_write_digit( ucDigit1, UNTRANSLATED_MODE, 0b01000000 );

    // display number on digit 2
    ucRc = saa1064_write_digit( ucDigit2, TRANSLATED_MODE, (unsigned char) (unsigned int) abs( iNumber ) );
  }
  else
  {
    if ( iNumber < 10 )
    {
      // blank out digit 1
      ucRc = saa1064_write_digit( ucDigit1, UNTRANSLATED_MODE, 0 );

      // display number in digit 2
      ucRc = saa1064_write_digit( ucDigit2, TRANSLATED_MODE, (unsigned char) (unsigned int) iNumber );
    }
    else
    {
      ucHigh = (unsigned char) (unsigned int) iNumber / 10;
      ucLow =  (unsigned char) (unsigned int) iNumber % 10;
      ucRc = saa1064_write_digit( ucDigit1, TRANSLATED_MODE, ucHigh );
      ucRc = saa1064_write_digit( ucDigit2, TRANSLATED_MODE, ucLow );
    }
  }

  return ucRc;
}

/*-----------------------------------------------------------------
  saa1064_write_digit(...)

  Input: which display
           1 = 1st (leftmost) digit
           2 = 2nd digit
           3 = 3rd digit
           4 = 4th (rightmost) digit
         translate
           0 = don't
           1 = do 
         digit 
           0..9 or 10..19 (translated mode)
           bits corresponding to segments to display (untranslated mode)

        The untranslated code passed in the digit fields tells the
        SAA1064 which segments to turn on i.e. P1..P8 or P9..P16.
        What actually turns on depends on how the segments have been
        wired up. Given the way this project has been wired, translating
        allows you to specify the digits 0..9 and the rest is taken care of.

        To activate decimal point in translated mode, add 10 to the digit value.

  returns: 0 = OK
           1 = START failed to access device
           2 = write pointer byte fail
           3 = write digit fail
           4 = bad parameter
----------------------------------------------------------------*/
unsigned char saa1064_write_digit( unsigned char ucWhichDigit, 
                                   unsigned char ucTranslate,
                                   unsigned char ucDigit )
{
  unsigned char ucRc, ucDig;

  // Validate the input data
  if ( ucWhichDigit < SAA1064_DIGIT_1_PTR || ucWhichDigit > SAA1064_DIGIT_4_PTR )
    return 4;

  if ( ucTranslate != TRANSLATED_MODE && ucTranslate != UNTRANSLATED_MODE )
    return 4;

  if ( ucTranslate == TRANSLATED_MODE )
    if ( ucDigit > 19 )
      return 4;

  // Access the bus and device, output the address
  ucRc = i2c_start( SAA1064_ADDRESS_W );
  if ( ucRc == 1 )             // fail to access device
  {
    return 1;
  }

  // Output the register pointer
  ucRc = i2c_write( ucWhichDigit );
  if ( ucRc == 1 )             // fail
  {
    i2c_stop();
    return 2;
  }

  if ( ucTranslate )
    ucDig = translate( ucDigit );
  else
    ucDig = ucDigit;

  // write 1st digit
  ucRc = i2c_write( ucDig );

  i2c_stop();

  if ( ucRc == 1 )             // fail
  {
    return 3;
  }

  return 0;
}

/*-----------------------------------------------------------------
  translate(...)

  Translate a decimal number (or number + 10 as specified above)
  into segment code used by SAA1064 as wired in this project.
----------------------------------------------------------------*/
static unsigned char translate( unsigned char ucValue )
{
  unsigned char ucDecimalReqd;
  unsigned char ucNumberReqd;
  unsigned char ucResult;

  static unsigned char ucSegTable[10] =
  {
    0b00111111, 0b00000110, 0b01011011, 0b01001111, 0b01100110,
    0b01101101, 0b01111100, 0b00000111, 0b01111111, 0b01100111
  };

  if ( ucValue > 9 )
  {
    ucDecimalReqd = 1;
    ucNumberReqd = ucValue - 10;
  }
  else
  {
    ucDecimalReqd = 0;
    ucNumberReqd = ucValue;
  }

  if ( ucNumberReqd > 9 )     // invalid input
    return 0;

  ucResult = ucSegTable[ucNumberReqd];

  if ( ucDecimalReqd )
    ucResult |= 0x80;

  return ucResult;
}
